!---------------------------------------------------------------------@
subroutine relax(uri,uu,glmch,dt,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
!
!     Calculate of stiff term
!
  use pram, only : imax, jmax, kmax, nv, ns, glmcr, cfl
  implicit none

  integer :: i, j, k
  integer :: is1, ie1, js1, je1, ks1, ke1

  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: glmch, dt, kappa, tmp1, tmp2
  real(8) :: vx, vy, vz, bx, by, bz, ex, ey, ez, sg, vsq, gf, &
             e0x, e0y, e0z, etx, ety, etz, ep, epx, epy, epz, &
             ettx, etty, ettz, ft, fp
!
!=====================================================================@
!
! Parameter for hyperbolic divergence cleaning
  tmp1=0.0001d0
  tmp2=cfl*5.0
 ! kappa=glmch/tmp1
  kappa=1.d0

  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
 
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        bx=uri(7,i,j,k)
        by=uri(8,i,j,k)
        bz=uri(9,i,j,k)
        ex=uri(10,i,j,k)
        ey=uri(11,i,j,k)
        ez=uri(12,i,j,k)
        sg=uri(16,i,j,k)

        vsq=vx**2+vy**2+vz**2
        gf=1./sqrt(1.-vsq)

        e0x=vz*by-vy*bz
        e0y=vx*bz-vz*bx
        e0z=vy*bx-vx*by

!        etx=ex-e0x
!        ety=ey-e0y
!        etz=ez-e0z

!        if(gf .gt. 1.0d-10) then
!          ep=etx*vx+ety*vy+etz*vz
          ep=ex*vx+ey*vy+ez*vz
!        else
!          ep=0.d0
!        endif

        epx=ep*vx
        epy=ep*vy
        epz=ep*vz

        etx=ex-epx
        ety=ey-epy
        etz=ez-epz

        ft=exp(-0.5d0*sg*gf*dt)
        fp=exp(-0.5d0*sg*dt/gf)

        uu(10,i,j,k)=e0x+(etx-e0x)*ft+epx*fp ! E_1
        uu(11,i,j,k)=e0y+(ety-e0y)*ft+epy*fp ! E_2
        uu(12,i,j,k)=e0z+(etz-e0z)*ft+epz*fp ! E_3

! psi, phi for hyperbolic divergence cleaning
        
        uu(14,i,j,k)=uu(14,i,j,k)*exp(-0.5d0*kappa*dt) !Phi
!        uu(14,i,j,k)=uu(14,i,j,k)*exp(-tmp2)
        uu(15,i,j,k)=uu(15,i,j,k)*exp(-0.5d0*kappa*dt) !Psi
!        uu(15,i,j,k)=uu(15,i,j,k)*exp(-tmp2)
! 

      enddo
    enddo
  enddo
   
  return
end subroutine relax
!
!---------------------------------------------------------------------@
subroutine relaxb(uri,uu,glmch,dt,nm0,is1,ie1,js1,je1,ks1,ke1)
!---------------------------------------------------------------------@
!
!  Calculate stiff term (second step)
!
  use pram, only : imax, jmax, kmax, nv, ns, glmcr, cfl
  implicit none

  integer :: i, j, k, nm0
  integer :: is1, ie1, js1, je1, ks1, ke1

  real(8) :: uri(ns,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: glmch, dt, kappa, tmp1, tmp2
  real(8) :: vx, vy, vz, bx, by, bz, ex, ey, ez, sg, vsq, gf, &
             e0x, e0y, e0z, etx, ety, etz, ep, epx, epy, epz, &
             ettx, etty, ettz, ft, fp
!
!=====================================================================@
!
! Parameter for hyperbolic divergence cleaning
!
  tmp1=0.0001d0
  tmp2=cfl*5.0
 ! kappa=glmch/tmp1
  kappa=1.d0

  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
 
        vx=uri(2,i,j,k)
        vy=uri(3,i,j,k)
        vz=uri(4,i,j,k)
        bx=uri(7,i,j,k)
        by=uri(8,i,j,k)
        bz=uri(9,i,j,k)
        ex=uu(10,i,j,k)
        ey=uu(11,i,j,k)
        ez=uu(12,i,j,k)
        sg=uri(16,i,j,k)

        vsq=vx**2+vy**2+vz**2
        gf=1./sqrt(1.-vsq)

        e0x=vz*by-vy*bz
        e0y=vx*bz-vz*bx
        e0z=vy*bx-vx*by

!        etx=ex-e0x
!        ety=ey-e0y
!        etz=ez-e0z

!        if(gf .gt. 1.0d-10) then
!          ep=etx*vx+ety*vy+etz*vz
          ep=ex*vx+ey*vy+ez*vz
!        else
!          ep=0.d0
!        endif

        epx=ep*vx
        epy=ep*vy
        epz=ep*vz

        etx=ex-epx
        ety=ey-epy
        etz=ez-epz

        ft=exp(-0.5d0*sg*gf*dt)
        fp=exp(-0.5d0*sg*dt/gf)

        uu(10,i,j,k)=e0x+(etx-e0x)*ft+epx*fp ! E_1
        uu(11,i,j,k)=e0y+(ety-e0y)*ft+epy*fp ! E_2
        uu(12,i,j,k)=e0z+(etz-e0z)*ft+epz*fp ! E_3

      enddo
    enddo
  enddo

  do k=ks1,ke1
    do j=js1,je1
      do i=is1,ie1
! psi, phi for hyperbolic divergence cleaning 
        
        uu(14,i,j,k)=uu(14,i,j,k)*exp(-0.5d0*kappa*dt) ! Phi
!        uu(14,i,j,k)=uu(14,i,j,k)*exp(-tmp2)
        uu(15,i,j,k)=uu(15,i,j,k)*exp(-0.5d0*kappa*dt) ! Psi
!        uu(15,i,j,k)=uu(15,i,j,k)*exp(-tmp2)
!
      enddo
    enddo
  enddo
   
  return
end subroutine relaxb
