!----------------------------------------------------------------------
  subroutine rk2fst(uu,uh,ww,hi,ho,hii,hij,hik,hoi,hoj,hok,nm0, &
                    akap1b,akap2b,akap3b,is1,ie1,js1,je1,ks1,ke1)
!----------------------------------------------------------------------
!
!      First step of timeadvance step
!
!      Variables
!       uu: conserved variables
!       ww: numerical flux
!       ho, hi: metric variables
!
  use pram, only : imax, jmax, kmax, nv
  implicit none
!
  integer :: i, j, k, m, nm0, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1), &
             uh(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: ww(3,nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: hi(3,nv,is1:ie1,js1:je1,ks1:ke1), &
             hii(nv,is1:ie1,js1:je1,ks1:ke1), & 
             hij(nv,is1:ie1,js1:je1,ks1:ke1), &
             hik(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: ho(3,nv,is1:ie1,js1:je1,ks1:ke1), &
             hoi(nv,is1:ie1,js1:je1,ks1:ke1), &
             hoj(nv,is1:ie1,js1:je1,ks1:ke1), &
             hok(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: akap1b(imax-1),akap2b(jmax-1),akap3b(kmax-1)

!  real(8) akap1a, akap2a, akap3a
      
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        do m=1,nv

          uh(m,i,j,k)=uu(m,i,j,k)+ &
        0.5*(akap1b(i)*ho(1,m,i,j,k) &
             *(hii(m,i-1,j,k)*ww(1,m,i-1,j,k)-hii(m,i,j,k)*ww(1,m,i,j,k)) &
           +akap2b(j)*ho(2,m,i,j,k) &
             *(hij(m,i,j-1,k)*ww(2,m,i,j-1,k)-hij(m,i,j,k)*ww(2,m,i,j,k)) &
           +akap3b(k)*ho(3,m,i,j,k) &
             *(hik(m,i,j,k-1)*ww(3,m,i,j,k-1)-hik(m,i,j,k)*ww(3,m,i,j,k)))
!
        enddo
      enddo
    enddo
  enddo
 
  return
end subroutine rk2fst
!
!----------------------------------------------------------------------
subroutine rk2snd(uu,ww,hi,ho,hii,hij,hik,hoi,hoj,hok,nm0, &
                  akap1b,akap2b,akap3b,is1,ie1,js1,je1,ks1,ke1)
!----------------------------------------------------------------------
!
!      Second step of timeadvance step
!
!      Variables
!       uu: conserved variables(new), 
!       ww: numerical flux
!       ho, hi: metric variables
!
  use pram, only : imax, jmax, kmax, nv
  implicit none
!
  integer :: i, j, k, m, nm0, is1, ie1, js1, je1, ks1, ke1

  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: ww(3,nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: hi(3,nv,is1:ie1,js1:je1,ks1:ke1), &
             hii(nv,is1:ie1,js1:je1,ks1:ke1), & 
             hij(nv,is1:ie1,js1:je1,ks1:ke1), &
             hik(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: ho(3,nv,is1:ie1,js1:je1,ks1:ke1), &
             hoi(nv,is1:ie1,js1:je1,ks1:ke1), &
             hoj(nv,is1:ie1,js1:je1,ks1:ke1), &
             hok(nv,is1:ie1,js1:je1,ks1:ke1)

  real(8) :: akap1b(imax-1),akap2b(jmax-1),akap3b(kmax-1)

!  real(8) akap1a, akap2a, akap3a
!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        do m=1,nv
!
          uu(m,i,j,k)=uu(m,i,j,k)+ &
           (akap1b(i)*ho(1,m,i,j,k) &
             *(hii(m,i-1,j,k)*ww(1,m,i-1,j,k)-hii(m,i,j,k)*ww(1,m,i,j,k)) &
            +akap2b(j)*ho(2,m,i,j,k) &
             *(hij(m,i,j-1,k)*ww(2,m,i,j-1,k)-hij(m,i,j,k)*ww(2,m,i,j,k)) &
            +akap3b(k)*ho(3,m,i,j,k) &
             *(hik(m,i,j,k-1)*ww(3,m,i,j,k-1)-hik(m,i,j,k)*ww(3,m,i,j,k)))
!  
        enddo
      enddo
    enddo
  enddo 
!
  return
end subroutine rk2snd
!
!----------------------------------------------------------------------
subroutine rk2adsff(uu,sf,nm0,is1,ie1,js1,je1,ks1,ke1)
!----------------------------------------------------------------------
!
!      Add source term to conservec variables for first step
!
!      Variables
!       uh: conserved variables
!       sf, sou: source term
!
  use pram, only : imax, jmax, kmax, nv
  implicit none
!
  integer :: i, j, k, nm0, is1, ie1, js1, je1, ks1, ke1
  
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: sf(nv,is1:ie1,js1:je1,ks1:ke1)

!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        uu(2,i,j,k)=uu(2,i,j,k)+sf(2,i,j,k)*0.5
        uu(3,i,j,k)=uu(3,i,j,k)+sf(3,i,j,k)*0.5
        uu(4,i,j,k)=uu(4,i,j,k)+sf(4,i,j,k)*0.5
        uu(5,i,j,k)=uu(5,i,j,k)+sf(5,i,j,k)*0.5
        uu(10,i,j,k)=uu(10,i,j,k)+sf(10,i,j,k)*0.5
        uu(11,i,j,k)=uu(11,i,j,k)+sf(11,i,j,k)*0.5
        uu(12,i,j,k)=uu(12,i,j,k)+sf(12,i,j,k)*0.5
        uu(15,i,j,k)=uu(15,i,j,k)+sf(15,i,j,k)*0.5  
      enddo
    enddo
  enddo
!
  return
end subroutine rk2adsff
!
!----------------------------------------------------------------------
subroutine rk2adsfs(uu,sf,nm0,is1,ie1,js1,je1,ks1,ke1)
!----------------------------------------------------------------------
!
!      Add source term to conservec variables for second step
!
!      Variables
!       uu: conserved variables
!       sf, sou: source term
!
  use pram, only : imax, jmax, kmax, nv
  implicit none
!
  integer :: i, j, k, nm0, is1, ie1, js1, je1, ks1, ke1
  
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: sf(nv,is1:ie1,js1:je1,ks1:ke1)
!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
          
        uu(2,i,j,k)=uu(2,i,j,k)+sf(2,i,j,k)
        uu(3,i,j,k)=uu(3,i,j,k)+sf(3,i,j,k)
        uu(4,i,j,k)=uu(4,i,j,k)+sf(4,i,j,k)
        uu(5,i,j,k)=uu(5,i,j,k)+sf(5,i,j,k)
        uu(10,i,j,k)=uu(10,i,j,k)+sf(10,i,j,k)
        uu(11,i,j,k)=uu(11,i,j,k)+sf(11,i,j,k)
        uu(12,i,j,k)=uu(12,i,j,k)+sf(12,i,j,k)
        uu(15,i,j,k)=uu(15,i,j,k)+sf(15,i,j,k) 
      enddo
    enddo
  enddo
!
  return
end subroutine rk2adsfs
!
!----------------------------------------------------------------------
subroutine rk_bf(uu,u0,nm0,is1,ie1,js1,je1,ks1,ke1)
!----------------------------------------------------------------------
!
!      Add initial conserved variables for time evolution of 
!      Magnetic field
!
  use pram, only : imax, jmax, kmax, nv
  implicit none
!
  integer :: i, j, k, nm0, is1, ie1, js1, je1, ks1, ke1
  
  real(8) :: uu(nv,is1:ie1,js1:je1,ks1:ke1)
  real(8) :: u0(nv,is1:ie1,js1:je1,ks1:ke1)

!
  do k=ks1+nm0,ke1-nm0
    do j=js1+nm0,je1-nm0
      do i=is1+nm0,ie1-nm0
        uu(7,i,j,k)=u0(7,i,j,k)+uu(7,i,j,k)
        uu(8,i,j,k)=u0(8,i,j,k)+uu(8,i,j,k)
        uu(9,i,j,k)=u0(9,i,j,k)+uu(9,i,j,k)
      enddo
    enddo
  enddo
!
  return
end subroutine rk_bf
